from pathlib import Path

import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go

from gtracr.constants import KG_M_S_PER_GEVC

PLOT_DIR = Path(__file__).parent.parent.parent / "gtracr_plots"

COLOR_LIST = ["b", "r", "c", "m", "y", "g", "k"]


def plot_3dtraj(
    trajectory_datalist,
    title_name="Particle Trajectory",
    file_name="test_trajectory_3d.html",
    mpl=False,
    plotdir_path=PLOT_DIR,
    show_plot=False,
):
    """
    Plots trajectories using a 3-dimensional plot using PlotLy, an interactive HTML plotting module. Options are available to plot in usual matplotlib as well.

    Parameters
    ----------

    - trajectory_datalist : list of dict(str, np.array)
        The list of dictionaries that stores the information of the trajectory in Cartesian coordinates.
    - title_name : str
        The title name of the plot. Default is set to `"Particle Trajectory"`.
    - file_name : str
        The name of the file that we want to save. This defaults to a filename with a .html extension. *The filename should not be .html if `mpl=True`.*
    - mpl : bool
        Enables plotting with matplotlib instead of PlotLy (default = False).
    - plotdir_path : str
        The path to the directory in which the plots are stored in. Default is set to a directory `gtracr_plots` placed in parallel with the root directory.
    - show_plot : bool
        Boolean whether to show the plot or not
    """

    # unpack dictionary
    data_list = []
    max_tarr_list = []

    for i, trajectory_data in enumerate(trajectory_datalist):
        data_list.append(
            (
                trajectory_data["t"],
                trajectory_data["x"],
                trajectory_data["y"],
                trajectory_data["z"],
            )
        )
        max_tarr_list.append(np.max(trajectory_data["t"]))

    # get maximal array of time for plotting purposes
    max_tarr_index = np.argmax(max_tarr_list, axis=0)
    cbar_tarr = data_list[max_tarr_index][0]

    print(
        f"Maximal and minimal time values: {np.max(cbar_tarr):.3e}, {np.min(cbar_tarr):.3e}"
    )

    # set the earth wireframe
    u, v = np.mgrid[
        0 : 2 * np.pi : 100j, 0 : np.pi : 100j
    ]  # x in a:b:xj is number of points in [a,b]
    x_sphere = np.sin(u) * np.cos(v)
    y_sphere = np.sin(u) * np.sin(v)
    z_sphere = np.cos(u)

    # plot using matplotlib
    if mpl:
        # set the figure and axes
        fig_3d = plt.figure(figsize=(12, 6))
        ax_3d = fig_3d.add_subplot(111, projection="3d")

        # plot the sphere
        ax_3d.plot_wireframe(x_sphere, y_sphere, z_sphere, color="k")

        # plot the trajectory
        for i, (t_arr, x_arr, y_arr, z_arr) in enumerate(data_list):
            cm_3d = ax_3d.scatter(x_arr, y_arr, z_arr, c=t_arr, marker="o", s=1.0)
            if i == max_tarr_index:
                cm3d_max = cm_3d

        # labels, colorbars, limits whatnot
        cbar_3d = fig_3d.colorbar(cm3d_max, ax=ax_3d)
        ax_3d.set_xlim([-10.0, 10.0])
        ax_3d.set_ylim([-10.0, 10.0])
        ax_3d.set_zlim([-10.0, 10.0])
        ax_3d.set_xlabel(r"x [$R_E$]", fontsize=12)
        ax_3d.set_ylabel(r"y [$R_E$]", fontsize=12)
        ax_3d.set_zlabel(r"z [$R_E$]", fontsize=12)
        cbar_3d.ax.set_ylabel("Time [s]", fontsize=12)
        ax_3d.set_title(title_name, fontsize=16)

        # make file extension to png if it is not png or jpg
        if file_name.find("png") < 0 or file_name.find("jpg") < 0:
            file_name = file_name.split(".")[0] + ".png"

        # mpld3.save_html(fig_3d, Path(PLOT_DIR) / "test_trajectory_3d.html")
        plt.savefig(Path(plotdir_path) / file_name)

        if show_plot:
            plt.show()

    # plot using PlotLy
    else:
        # construct trajectory plot object
        traj_plots = []
        for i, (t_arr, x_arr, y_arr, z_arr) in enumerate(data_list):
            # marker setings for trajectory plot

            traj_colorbar = (
                dict(thickness=20, title="Time [s]") if i == max_tarr_index else None
            )
            traj_marker = dict(
                size=4,
                color=t_arr,  # set color to an array/list of desired values
                colorscale="Viridis",  # choose a colorscale
                opacity=0.8,
                colorbar=traj_colorbar,
            )

            traj_plot = go.Scatter3d(
                x=x_arr, y=y_arr, z=z_arr, mode="markers", marker=traj_marker
            )
            traj_plots.append(traj_plot)

        # construct wireframe for sphere
        lines = []
        line_marker = dict(color="#0066FF", width=2)
        for i, j, k in zip(x_sphere, y_sphere, z_sphere):
            lines.append(go.Scatter3d(x=i, y=j, z=k, mode="lines", line=line_marker))

        # append them all and plot
        data = lines + traj_plots
        fig = go.Figure(data=data)

        # additional configurations
        # somehow latex axis labels are still not enabled with plotly?
        fig.update_layout(
            margin=dict(l=0, r=0, b=0, t=0),
            scene_aspectmode="cube",
            scene=dict(
                xaxis=dict(nticks=6, range=[-10.0, 10.0], title=r"x [Re]"),
                yaxis=dict(nticks=6, range=[-10.0, 10.0], title=r"y [Re]"),
                zaxis=dict(nticks=6, range=[-10.0, 10.0], title=r"z [Re]"),
            ),
            font=dict(family="Courier New, monospace", size=18),
            showlegend=False,
        )

        # make file extension to html if it is not html
        if file_name.find("html") < 0:
            file_name = file_name.split(".")[0] + ".html"

        if show_plot:
            fig.show()

        # write to html
        fig.write_html(Path(plotdir_path) / file_name)


def plot_2dtraj(
    trajectory_datalist,
    dim1="x",
    dim2="y",
    title_name="Particle Trajectory",
    file_name="test_trajectory_proj.png",
    mpl=False,
    plotdir_path=PLOT_DIR,
    show_plot=False,
):
    """
    Plots the projections of the trajectory in three different planes, and the time evolution of the magnitude of the trajectory. Only supported with matplotlib for now.

    Parameters
    ----------

    - trajectory_datalist : list of dict(str, np.array)
        The list of dictionaries that stores the information of the trajectory in Cartesian coordinates.
    - dim1, dim2 : str
        The two dimensions in which we want to plot the trajectories on.
    - title_name : str
        The title name of the plot. Default is setot `"Particle Trajectory"`.
    - file_name : str
        The name of the file that we want to save. This defaults to a filename with a .html extension. *The filename should not be .html if `mpl=True`.*
    - mpl : bool
        Enables plotting with matplotlib instead of PlotLy (default = False).
    - plotdir_path : str
        The path to the directory in which the plots are stored in. Default is set to a directory `gtracr_plots` placed in parallel with the root directory.
    - show_plot : bool
        Boolean whether to show the plot or not
    """
    data_list = []
    max_tarr_list = []

    for i, trajectory_data in enumerate(trajectory_datalist):
        data_list.append(
            (trajectory_data["t"], trajectory_data[dim1], trajectory_data[dim2])
        )
        max_tarr_list.append(np.max(trajectory_data["t"]))

    # get maximal array of time for plotting purposes
    max_tarr_index = np.argmax(max_tarr_list, axis=0)
    cbar_tarr = data_list[max_tarr_index][0]

    print(
        f"Maximal and minimal time values: {np.max(cbar_tarr):.3e}, {np.min(cbar_tarr):.3e}"
    )

    fig_2d, ax_2d = plt.subplots(figsize=(12, 9), constrained_layout=True)

    for i, (t_arr, dim1_arr, dim2_arr) in enumerate(data_list):
        plot_colorbar = True if i == max_tarr_index else False
        plot_traj_projection(
            dim1_arr, dim2_arr, t_arr, fig_2d, ax_2d, dim1, dim2, plot_colorbar
        )

    # fig_proj.suptitle(title_name, fontsize=16)
    if show_plot:
        plt.show()
    # html doesnt work with latex labels and suptitle, so its deprecated for now
    # mpld3.save_html(fig_proj,
    #                 Path(PLOT_DIR) / "test_trajectory_proj.html")
    plt.savefig(Path(PLOT_DIR) / "test_trajectory_proj.png")


def plot_traj_projection(arr1, arr2, t_arr, fig, ax, label1, label2, plot_colorbar):
    """
    Plot the projection of the trajectory.

    Parameters
    ----------
    arr1 (numpy array):
        the array that goes on the x-axis
    arr2 (numpy array):
        the array that goes on the y-axis
    t_arr (numpy array):
        the time array
    fig, ax:
        matplotlib plt.Figure and plt.Axes.axes objects
    label1, label2 (str):
        the labels that indicate the identity of arr1 and arr2 resp.
    plot_colorbar (bool):
        boolean that decides whether to plot colorbar or not
    """
    cm = ax.scatter(arr1, arr2, c=t_arr)
    circ = patches.Circle(
        (0.0, 0.0), 1.0, alpha=0.8, fc="#1f77b4", linestyle="-", ec="b", lw=4.0
    )
    ax.add_patch(circ)

    ax.set_xlim([-10, 10])
    ax.set_ylim([-10, 10])
    ax.set_xlabel(rf"{label1:s} [$R_E$]", fontsize=17)
    ax.set_ylabel(rf"{label2:s} [$R_E$]", fontsize=17)

    ax.tick_params(axis="x", labelsize=17)
    ax.tick_params(axis="y", labelsize=17)

    if plot_colorbar:
        cbar = fig.colorbar(cm, ax=ax)
        cbar.ax.set_ylabel("Time [s]", fontsize=17)
        cbar.ax.tick_params(axis="y", labelsize=17)
    ax.set_title(f"Trajectory projected onto {label1:s}-{label2:s} plane", fontsize=17)


def plot_traj_momentum(trajectory_data, p0, show_plot=False):
    """
    Plot the time evolution of the magnitude of the momentum.
    This is mainly used for debugging purposes as a cross-check
    to the accuracy of our trajectory simulation, as |p| should
    be constant throughout the trajectory.

    Parameters
    -----------

    - trajectory_data : dict(str, np.array)
        The dictionary that contains the momentum components of the
        trajectory.
    - p0 : float
        The initial momentum of the trajectory
    - show_plot : bool
        Decide whether to show the plot in a GUI or not (default = False).
    """
    # check momentum magnitude vs steps since |p| should be
    # constant throughout the trajectory
    t_arr = trajectory_data["t"]
    p_arr = (
        np.sqrt(
            trajectory_data["pr"] ** 2.0
            + trajectory_data["ptheta"] ** 2.0
            + trajectory_data["pphi"] ** 2.0
        )
        / KG_M_S_PER_GEVC
    )
    p_ratio = p_arr / p0

    # figure for momentum vs steps
    # for physics checking
    fig_pmag, ax_pmag = plt.subplots(figsize=(12, 6), constrained_layout=True)

    # momentum ratio vs steps
    ax_pmag.plot(t_arr, p_ratio, color="b", marker="o", ms=3.0)
    # ax_pmag.set_ylim([0.5, 1.5])
    # ax_pmag.set_ylim([-3, 3])
    ax_pmag.set_xlabel(r"Time [s]", fontsize=14)
    ax_pmag.set_ylabel(r"$p/p_0$", fontsize=14)
    ax_pmag.set_title(
        r"Time Variation of Momentum Magnnitude Throughout Trajectory", fontsize=15
    )

    if show_plot:
        plt.show()
    plt.savefig(Path(PLOT_DIR) / "pmag_plot.png")


def plot_gmrc_scatter(
    gmrc_data,
    locname,
    plabel,
    bfield_type,
    iter_num,
    show_plot=False,
    plotdir_path=PLOT_DIR,
):
    """
    Plot the scatter plot of the geomagnetic rigidity cutoffs at the specified location and type of particle that the cosmic ray constitutes of. Currently only supports matplotlib.

    Parameters
    ----------
    - gmrc_data : dict(str, np.array)
        A dictionary that consists of the azimuthal and zenith components of the particle trajectory and its corresponding rigidity cutoff in the following order: (azimuth, zenith, rigidity cutoff)

    - locname : str
        The name of the detector location.

    - plabel : str
        The label of the particle that constitutes the cosmic ray.

    - show_plot : bool
        Decides to choose to show the generated plot or not. If True, presents the plot in a GUI window.

    - plotdir_path : str
        The path to the directory in which the plots are stored in. Default is set to a directory `gtracr_plots` placed in parallel with the root directory.

    """

    # get azimuth, zenith, and rigidity cutoff arrays
    azimuth_arr = gmrc_data["azimuth"]
    zenith_arr = gmrc_data["zenith"]
    rcutoff_arr = gmrc_data["rcutoff"]

    fig, ax = plt.subplots(figsize=(12, 6), constrained_layout=True)
    sc = ax.scatter(azimuth_arr, zenith_arr, c=rcutoff_arr, s=3.0)
    ax.set_xlabel("Azimuthal Angle [Degrees]")
    ax.set_ylabel("Zenith Angle [Degrees]")
    ax.set_title(
        f"Geomagnetic Rigidity Cutoffs at {locname} for {plabel} with N = {iter_num}"
    )

    cbar = fig.colorbar(sc, ax=ax)
    cbar.ax.set_ylabel("Rigidity [GV]")

    ax.set_xlim([0.0, 360.0])
    ax.set_ylim([180.0, 0.0])

    plt.savefig(
        Path(plotdir_path) / f"{locname}_{plabel}_{bfield_type}_scatterplot.png",
        dpi=800,
    )

    if show_plot:
        plt.show()


def plot_gmrc_heatmap(
    gmrc_grids,
    rigidity_list,
    locname,
    plabel,
    bfield_type,
    show_plot=False,
    plotdir_path=PLOT_DIR,
):
    """
    Plot the heatmap of the geomagnetic rigidity cutoffs at the specified
    location and particle type.

    Works with both the legacy ``interpolate_results`` output (irregularly
    spaced grids fed to ``imshow``) and the new ``bin_results`` output
    (regular bin-centre arrays fed to ``pcolormesh`` + ``contour``).

    Parameters
    ----------
    - gmrc_grids : tuple(np.array(float))
        (azimuth_grid, zenith_grid, rcutoff_grid)

    - rigidity_list : np.array(float)
        Rigidities evaluated in the MC (for colorbar limits / contour levels).

    - locname, plabel, bfield_type : str
    - show_plot : bool
    - plotdir_path : str
    """

    (azimuth_grid, zenith_grid, rcutoff_grid) = gmrc_grids
    vmin, vmax = np.min(rigidity_list), np.max(rigidity_list)

    fig, ax = plt.subplots(figsize=(12, 6), constrained_layout=True)

    # pcolormesh handles NaN bins gracefully (transparent), and works
    # directly with bin-centre arrays from bin_results().
    mesh = ax.pcolormesh(
        azimuth_grid,
        zenith_grid,
        rcutoff_grid,
        cmap="RdBu_r",
        vmin=vmin,
        vmax=vmax,
        shading="nearest",
        alpha=0.8,
    )

    # Contour lines over the heatmap.  Mask NaN so contour doesn't choke.
    masked_grid = np.ma.masked_invalid(rcutoff_grid)
    # nlevels = max(4, int(4 * len(rigidity_list) / 6))
    nlevels = [2.0, 5.0, 10.0, 20.0, 30.0, 40.0, 50.0, 100.0]
    cs = ax.contour(
        azimuth_grid,
        zenith_grid,
        masked_grid,
        colors="k",
        linewidths=1.0,
        levels=nlevels,
    )
    ax.clabel(cs, inline=True, fontsize=10, fmt="%.0f")

    cbar = fig.colorbar(mesh, ax=ax, shrink=0.8)
    cbar.ax.set_ylabel("Rigidity [GV]")

    # ylim from 180 to 0 to follow convention in Honda 2002 paper
    ax.set_xlim([0.0, 360.0])
    ax.set_ylim([180.0, 0.0])

    ax.set_xlabel("Azimuthal Angle [Degrees]")
    ax.set_ylabel("Zenith Angle [Degrees]")
    ax.set_title(f"Geomagnetic Rigidity Cutoffs at {locname} for {plabel}")

    if show_plot:
        plt.show()

    plt.savefig(Path(plotdir_path) / f"{locname}_{plabel}_{bfield_type}_cutoffplot.png")



"""
gtracr — cosmic ray trajectory simulation through Earth's geomagnetic field.

Simulates cosmic ray trajectories through Earth's geomagnetic field using
the IGRF-13 model and computes geomagnetic rigidity cutoffs (GMRC) via
Monte Carlo sampling.

Main classes
------------
Trajectory : Single cosmic ray trajectory evaluation.
GMRC : Geomagnetic rigidity cutoff map evaluation.
"""

from importlib.metadata import version

from gtracr.geomagnetic_cutoffs import GMRC
from gtracr.trajectory import Trajectory

__version__ = version("gtracr")

__all__ = ["Trajectory", "GMRC", "__version__"]



"""Tests for the pure-Python fallback trajectory tracer (pTrajectoryTracer)."""

import numpy as np
import pytest

from gtracr._fallback import pTrajectoryTracer
from gtracr.constants import (
    EARTH_RADIUS,
    ELEMENTARY_CHARGE,
    KG_M_S_PER_GEVC,
    KG_PER_GEVC2,
)

_DATA_DIR = str(
    (
        __import__("pathlib").Path(__file__).parent.parent / "src" / "gtracr" / "data"
    ).resolve()
)
_IGRF_PARAMS = (_DATA_DIR, 2020.0)

# pTrajectoryTracer uses start_altitude as altitude above Earth's surface (metres).
_START_ALT = 100e3  # 100 km in metres

# A proton at 100 km altitude, equatorial, with moderate momentum
_VEC0 = np.array([EARTH_RADIUS + _START_ALT, np.pi / 2.0, 0.0, 0.0, 0.0, 5.36e-19])


def _make_tracer(bfield_type="d", max_step=500):
    return pTrajectoryTracer(
        charge=ELEMENTARY_CHARGE,
        mass=0.938 * KG_PER_GEVC2,
        start_altitude=_START_ALT,
        escape_radius=10.0 * EARTH_RADIUS,
        stepsize=1e-4,
        max_step=max_step,
        bfield_type=bfield_type,
        igrf_params=_IGRF_PARAMS,
    )


def test_ptracer_dipole_evaluate():
    """evaluate() runs without error and sets final_time and final_sixvector."""
    tt = _make_tracer("d")
    tt.evaluate(0.0, _VEC0.copy())
    assert tt.final_time > 0.0
    assert len(tt.final_sixvector) == 6
    assert all(np.isfinite(v) for v in tt.final_sixvector)


def test_ptracer_igrf_evaluate():
    """evaluate() works with bfield_type='i' (IGRF-13)."""
    tt = _make_tracer("i")
    tt.evaluate(0.0, _VEC0.copy())
    assert tt.final_time > 0.0


def test_ptracer_dipole_get_trajectory():
    """evaluate_and_get_trajectory() returns dict with all required keys."""
    tt = _make_tracer("d")
    data = tt.evaluate_and_get_trajectory(0.0, _VEC0.copy())
    for key in ["t", "r", "theta", "phi", "pr", "ptheta", "pphi"]:
        assert key in data
    assert len(data["t"]) > 0


def test_ptracer_escaped_branch():
    """Radially outward high-momentum proton escapes (exercises lines 156-157)."""
    # Radial outward momentum at ~0.9998c guarantees escape within max_step steps.
    mom_si = 50.0 * KG_M_S_PER_GEVC
    vec0 = np.array([EARTH_RADIUS + _START_ALT, np.pi / 2.0, 0.0, mom_si, 0.0, 0.0])
    tt = pTrajectoryTracer(
        charge=ELEMENTARY_CHARGE,
        mass=0.938 * KG_PER_GEVC2,
        start_altitude=_START_ALT,
        escape_radius=10.0 * EARTH_RADIUS,
        stepsize=1e-5,
        max_step=200000,
        bfield_type="d",
    )
    tt.evaluate(0.0, vec0)
    assert tt.particle_escaped is True


def test_ptracer_forbidden_branch_evaluate():
    """Low-rigidity trajectory terminates at atmosphere (line 159 in evaluate)."""
    # Near-zero momentum: particle barely moves, will be pulled back inward.
    mom_si = 1e-22
    vec0 = np.array([EARTH_RADIUS + _START_ALT, np.pi / 2.0, 0.0, 0.0, 0.0, mom_si])
    tt = pTrajectoryTracer(
        charge=ELEMENTARY_CHARGE,
        mass=0.938 * KG_PER_GEVC2,
        start_altitude=_START_ALT,
        escape_radius=10.0 * EARTH_RADIUS,
        stepsize=1e-5,
        max_step=200000,
        bfield_type="d",
    )
    tt.evaluate(0.0, vec0)
    assert tt.particle_escaped is False


def test_ptracer_forbidden_branch_get_trajectory():
    """Forbidden trajectory in evaluate_and_get_trajectory() (line 199)."""
    mom_si = 1e-22
    vec0 = np.array([EARTH_RADIUS + _START_ALT, np.pi / 2.0, 0.0, 0.0, 0.0, mom_si])
    tt = pTrajectoryTracer(
        charge=ELEMENTARY_CHARGE,
        mass=0.938 * KG_PER_GEVC2,
        start_altitude=_START_ALT,
        escape_radius=10.0 * EARTH_RADIUS,
        stepsize=1e-5,
        max_step=200000,
        bfield_type="d",
    )
    data = tt.evaluate_and_get_trajectory(0.0, vec0)
    assert tt.particle_escaped is False
    assert len(data["t"]) > 0


def test_ptracer_escaped_branch_get_trajectory():
    """Escaped trajectory in evaluate_and_get_trajectory() (exercises lines 196-197)."""
    mom_si = 50.0 * KG_M_S_PER_GEVC
    vec0 = np.array([EARTH_RADIUS + _START_ALT, np.pi / 2.0, 0.0, mom_si, 0.0, 0.0])
    tt = pTrajectoryTracer(
        charge=ELEMENTARY_CHARGE,
        mass=0.938 * KG_PER_GEVC2,
        start_altitude=_START_ALT,
        escape_radius=10.0 * EARTH_RADIUS,
        stepsize=1e-5,
        max_step=200000,
        bfield_type="d",
    )
    data = tt.evaluate_and_get_trajectory(0.0, vec0)
    assert tt.particle_escaped is True
    assert len(data["t"]) > 0


def test_ptracer_invalid_bfield_raises():
    """Unknown bfield_type raises ValueError."""
    with pytest.raises(ValueError):
        pTrajectoryTracer(
            charge=ELEMENTARY_CHARGE,
            mass=0.938 * KG_PER_GEVC2,
            bfield_type="x",
        )


def test_ptracer_igrf_missing_params_raises():
    """bfield_type='i' without igrf_params raises ValueError."""
    with pytest.raises(ValueError, match="igrf_params"):
        pTrajectoryTracer(
            charge=ELEMENTARY_CHARGE,
            mass=0.938 * KG_PER_GEVC2,
            bfield_type="i",
            igrf_params=None,
        )


def test_trajectory_use_python_flag():
    """Trajectory.get_trajectory(use_python=True) exercises pTrajectoryTracer."""
    from gtracr.trajectory import Trajectory

    # Use a fixed IGRF-13 date (within its range of 1900-2025).
    traj = Trajectory(
        zenith_angle=45.0,
        azimuth_angle=90.0,
        rigidity=10.0,
        location_name="Kamioka",
        bfield_type="igrf",
        date="2020-06-15",
    )
    traj.get_trajectory(dt=1e-4, max_time=0.05, use_python=True)
    assert traj.final_time > 0.0



"""
Tests for Python B-field wrappers: MagneticField (dipole), IGRF13, and IGRFTable.
"""

from pathlib import Path

import numpy as np
import pytest

from gtracr.constants import EARTH_RADIUS

DATA_DIR = Path(__file__).parent.parent / "src" / "gtracr" / "data"


# ---------------------------------------------------------------------------
# MagneticField (dipole) — Python wrapper
# ---------------------------------------------------------------------------


def test_dipole_equator_finite():
    """MagneticField at equator (theta=pi/2) returns finite values."""
    from gtracr.bfield.dipole import MagneticField

    bf = MagneticField()
    result = bf.values(EARTH_RADIUS, np.pi / 2.0, 0.0)
    assert len(result) == 3
    assert all(np.isfinite(v) for v in result)


def test_dipole_pole_finite():
    """MagneticField at north pole (theta≈0) returns finite values."""
    from gtracr.bfield.dipole import MagneticField

    bf = MagneticField()
    result = bf.values(EARTH_RADIUS, 1e-9, 0.0)
    assert len(result) == 3
    assert all(np.isfinite(v) for v in result)


def test_dipole_analytical():
    """Spot-check: dipole B at equator (theta=pi/2) — Br should be zero."""
    from gtracr.bfield.dipole import MagneticField

    bf = MagneticField()
    br, bth, bph = bf.values(EARTH_RADIUS, np.pi / 2.0, 0.0)
    # At equator the radial component of a dipole is zero.
    # Use atol=1e-15 T to accommodate floating-point error in cos(pi/2).
    assert np.isclose(br, 0.0, atol=1e-15)


def test_dipole_r_scaling():
    """Dipole B magnitude scales as 1/r^3."""
    from gtracr.bfield.dipole import MagneticField

    bf = MagneticField()
    b1 = np.linalg.norm(bf.values(EARTH_RADIUS, np.pi / 2.0, 0.0))
    b2 = np.linalg.norm(bf.values(2.0 * EARTH_RADIUS, np.pi / 2.0, 0.0))
    ratio = b1 / b2
    assert np.isclose(ratio, 8.0, rtol=1e-6)  # (2RE/RE)^3 = 8


# ---------------------------------------------------------------------------
# IGRF13 — Python wrapper
# ---------------------------------------------------------------------------


def test_igrf13_default_nmax():
    """IGRF13 default nmax is 13."""
    from gtracr.bfield.igrf import IGRF13

    igrf = IGRF13(2020.0)
    assert igrf.nmax == 13


def test_igrf13_custom_nmax():
    """IGRF13 with nmax=5 initializes without error."""
    from gtracr.bfield.igrf import IGRF13

    igrf = IGRF13(2020.0, nmax=5)
    assert igrf.nmax == 5
    result = igrf.values(EARTH_RADIUS, np.pi / 2.0, 0.0)
    assert all(np.isfinite(v) for v in result)


@pytest.mark.parametrize("year", [1900, 1950, 2020, 2024])
def test_igrf13_date_parametrize(year):
    """IGRF13 returns finite values across all epochs."""
    from gtracr.bfield.igrf import IGRF13

    igrf = IGRF13(float(year))
    result = igrf.values(EARTH_RADIUS, np.pi / 2.0, np.pi)
    assert all(np.isfinite(v) for v in result)


def test_igrf13_values_finite():
    """Spot-check multiple (r, theta, phi) coordinates."""
    from gtracr.bfield.igrf import IGRF13

    igrf = IGRF13(2020.0)
    coords = [
        (EARTH_RADIUS, np.pi / 2.0, 0.0),
        (2.0 * EARTH_RADIUS, 1.0, np.pi / 4.0),
        (5.0 * EARTH_RADIUS, np.pi / 3.0, np.pi),
    ]
    for r, theta, phi in coords:
        result = igrf.values(r, theta, phi)
        assert all(np.isfinite(v) for v in result), (
            f"Non-finite at ({r}, {theta}, {phi})"
        )


def test_igrf13_leap_year():
    """IGRF13 with a leap year initializes correctly."""
    from gtracr.bfield.igrf import IGRF13

    igrf = IGRF13(2000)  # 2000 is a leap year
    result = igrf.values(EARTH_RADIUS, np.pi / 2.0, np.pi)
    assert len(result) == 3


def test_igrf13_large_theta_normalization():
    """IGRF13 normalizes theta_deg > 180 without crashing."""
    from gtracr.bfield.igrf import IGRF13

    igrf = IGRF13(2015)
    # theta_rad = 4.0 → ~229 degrees > 180, triggers normalization
    result = igrf.values(EARTH_RADIUS, 4.0, np.pi)
    assert len(result) == 3


def test_igrf13_large_phi_normalization():
    """IGRF13 normalizes phi_deg > 360 without crashing."""
    from gtracr.bfield.igrf import IGRF13

    igrf = IGRF13(2015)
    # phi_rad = 7.5 → ~430 degrees > 360, triggers normalization
    result = igrf.values(EARTH_RADIUS, np.pi / 2.0, 7.5)
    assert len(result) == 3


# ---------------------------------------------------------------------------
# IGRFTable — Python lookup table wrapper
# ---------------------------------------------------------------------------


def test_igrf_table_values_finite():
    """IGRFTable.values() returns finite floats at a sample point."""
    from gtracr.bfield.table import NPHI, NR, NTHETA, IGRFTable

    # Build a constant table manually to avoid running the full build loop
    table = IGRFTable.__new__(IGRFTable)
    table._r_min = EARTH_RADIUS
    table._r_max = 10.0 * EARTH_RADIUS
    table._log_r_min = np.log(EARTH_RADIUS)
    table._log_r_max = np.log(10.0 * EARTH_RADIUS)
    table._table = np.ones((3, NR, NTHETA, NPHI), dtype=np.float32) * 1e-5

    result = table.values(EARTH_RADIUS, np.pi / 2.0, 0.0)
    assert len(result) == 3
    assert all(np.isfinite(v) for v in result)


def test_igrf_table_phi_periodicity():
    """IGRFTable.values() at phi=0 and phi=2*pi returns the same result."""
    from gtracr.bfield.table import NPHI, NR, NTHETA, IGRFTable

    table = IGRFTable.__new__(IGRFTable)
    table._r_min = EARTH_RADIUS
    table._r_max = 10.0 * EARTH_RADIUS
    table._log_r_min = np.log(EARTH_RADIUS)
    table._log_r_max = np.log(10.0 * EARTH_RADIUS)
    # Set a non-uniform table so interpolation results depend on phi
    rng = np.random.default_rng(42)
    table._table = rng.random((3, NR, NTHETA, NPHI)).astype(np.float32)

    r = EARTH_RADIUS * 2.0
    theta = np.pi / 2.0
    result0 = table.values(r, theta, 0.0)
    result2pi = table.values(r, theta, 2.0 * np.pi)
    # phi=0 and phi=2pi should map to the same cell (ip0=0 in both cases)
    for v0, v2 in zip(result0, result2pi):
        assert np.isclose(v0, v2, rtol=1e-5)


def test_igrf_table_shape_constant():
    """IGRFTable NR, NTHETA, NPHI match expected C++ grid dimensions."""
    from gtracr.bfield.table import NPHI, NR, NTHETA

    assert NR == 64
    assert NTHETA == 128
    assert NPHI == 256


def test_igrf_table_r_clamping():
    """IGRFTable.values() clamps r below r_min without crashing."""
    from gtracr.bfield.table import NPHI, NR, NTHETA, IGRFTable

    table = IGRFTable.__new__(IGRFTable)
    table._r_min = EARTH_RADIUS
    table._r_max = 10.0 * EARTH_RADIUS
    table._log_r_min = np.log(EARTH_RADIUS)
    table._log_r_max = np.log(10.0 * EARTH_RADIUS)
    table._table = np.ones((3, NR, NTHETA, NPHI), dtype=np.float32)

    # r below r_min should be clamped
    result = table.values(EARTH_RADIUS * 0.5, np.pi / 2.0, 0.0)
    assert all(np.isfinite(v) for v in result)


def test_igrf_table_r_max_clamping():
    """IGRFTable.values() clamps r above r_max without crashing."""
    from gtracr.bfield.table import NPHI, NR, NTHETA, IGRFTable

    table = IGRFTable.__new__(IGRFTable)
    table._r_min = EARTH_RADIUS
    table._r_max = 10.0 * EARTH_RADIUS
    table._log_r_min = np.log(EARTH_RADIUS)
    table._log_r_max = np.log(10.0 * EARTH_RADIUS)
    table._table = np.ones((3, NR, NTHETA, NPHI), dtype=np.float32)

    # r above r_max should be clamped
    result = table.values(EARTH_RADIUS * 20.0, np.pi / 2.0, 0.0)
    assert all(np.isfinite(v) for v in result)



"""Geomagnetic rigidity cutoff evaluation via Monte Carlo sampling."""

import threading
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from datetime import date
from pathlib import Path

import numpy as np
from scipy.interpolate import griddata
from tqdm import tqdm

from gtracr._libgtracr import TrajectoryTracer as CppTrajectoryTracer
from gtracr.constants import (
    EARTH_RADIUS,
    ELEMENTARY_CHARGE,
    KG_M_S_PER_GEVC,
    KG_PER_GEVC2,
    SOLVER_CHARS,
)
from gtracr.trajectory import Trajectory

_DATA_DIR = Path(__file__).parent / "data"

_thread_local = threading.local()

# ---------------------------------------------------------------------------
# Process-level TrajectoryTracer cache (for ProcessPoolExecutor workers)
# ---------------------------------------------------------------------------
# Each worker process stores one shared CppTrajectoryTracer here.  It is
# initialised once by _init_worker() (via ProcessPoolExecutor's initializer
# argument) so the expensive IGRF JSON load happens only *once per worker
# process* rather than once per trajectory direction.
_worker_tracer = None


def _init_worker(common_params):
    """Initialise a single CppTrajectoryTracer for the lifetime of this worker.

    Called automatically by ProcessPoolExecutor when a new worker process
    starts.  Storing the tracer in a module-level variable avoids the IGRF
    JSON parse for every MC direction handled by that worker.

    Parameters
    ----------
    common_params : tuple
        ``(plabel, bfield_type_char, date_str, palt, dt, max_time,
           solver_char, atol, rtol)``
    """
    global _worker_tracer

    plabel, bfield_type_char, date_str, palt, dt, max_time, solver_char, atol, rtol = (
        common_params
    )

    from gtracr.utils import particle_dict, ymd_to_dec

    particle = particle_dict[plabel]
    particle.set_from_rigidity(10.0)  # arbitrary rigidity to fix charge/mass
    charge_si = particle.charge * ELEMENTARY_CHARGE
    mass_si = particle.mass * KG_PER_GEVC2

    datapath = str(_DATA_DIR.resolve())
    dec_date = float(ymd_to_dec(date_str))
    igrf_params = (datapath, dec_date)

    # Use a placeholder start_alt; it is updated per direction via
    # set_start_altitude() before each evaluate() call.
    nominal_start_alt = palt * 1e3 + EARTH_RADIUS
    esc_alt = 10.0 * EARTH_RADIUS
    max_step = int(np.ceil(max_time / dt))

    _worker_tracer = CppTrajectoryTracer(
        charge_si,
        mass_si,
        nominal_start_alt,
        esc_alt,
        dt,
        max_step,
        bfield_type_char,
        igrf_params,
        solver_char,
        atol,
        rtol,
    )


# Directory for cached IGRF tables (next to the data files).
_TABLE_CACHE_DIR = _DATA_DIR


def _get_or_generate_igrf_table(datapath, dec_date):
    """Load a cached IGRF table from disk, or generate and cache it.

    The table depends on the decimal year and the compile-time grid
    constants (Nr, Ntheta, Nphi).  The cache files (.npy for the table,
    .npz for the params) are stored alongside igrf13.json.
    """
    from gtracr._libgtracr import TableParams
    from gtracr._libgtracr import generate_igrf_table as _gen_table

    cache_table = _TABLE_CACHE_DIR / f"igrf_table_{dec_date:.4f}.npy"
    cache_params = _TABLE_CACHE_DIR / f"igrf_table_{dec_date:.4f}_params.npz"

    if cache_table.is_file() and cache_params.is_file():
        table_flat = np.load(str(cache_table))
        d = np.load(str(cache_params))
        table_params = TableParams()
        table_params.r_min = float(d["r_min"])
        table_params.r_max = float(d["r_max"])
        table_params.log_r_min = float(d["log_r_min"])
        table_params.log_r_max = float(d["log_r_max"])
        table_params.Nr = int(d["Nr"])
        table_params.Ntheta = int(d["Ntheta"])
        table_params.Nphi = int(d["Nphi"])
        return table_flat, table_params

    # Generate from scratch and cache.
    table_flat, table_params = _gen_table(datapath, dec_date)
    np.save(str(cache_table), table_flat)
    np.savez(
        str(cache_params),
        r_min=table_params.r_min,
        r_max=table_params.r_max,
        log_r_min=table_params.log_r_min,
        log_r_max=table_params.log_r_max,
        Nr=table_params.Nr,
        Ntheta=table_params.Ntheta,
        Nphi=table_params.Nphi,
    )
    return table_flat, table_params


def _evaluate_single_direction(args):
    """
    Evaluate the cutoff rigidity for a single random (zenith, azimuth) direction.

    Two levels of TrajectoryTracer caching are used:

    1. **Process-level cache**: when the worker was initialised via
       ``_init_worker``, the module-level ``_worker_tracer`` is reused across
       all MC samples in that worker — only ``set_start_altitude()`` and
       ``reset()`` are called between samples, so the IGRF is loaded just once
       per CPU core.

    2. **Per-direction fallback**: when no process-level tracer exists (e.g.
       ``n_workers=1``), a single CppTrajectoryTracer is created for the
       current direction and reused across the rigidity scan.

    Parameters
    ----------
    args : tuple
        (location, plabel, bfield_type, date_str, palt, rigidity_list, dt, max_time,
         seed, solver_char, atol, rtol)

    Returns
    -------
    (azimuth, zenith, rcutoff) or (azimuth, zenith, 0.0) if no cutoff found
    """
    (
        location,
        plabel,
        bfield_type,
        date_str,
        palt,
        rigidity_list,
        dt,
        max_time,
        seed,
        solver_char,
        atol,
        rtol,
    ) = args
    rng = np.random.default_rng(seed)
    azimuth, zenith = rng.random(2) * np.array([360.0, 180.0])

    # Build Trajectory once for geometry; position and momentum direction are
    # the same for all rigidities — only momentum magnitude changes.
    traj = Trajectory(
        plabel=plabel,
        location_name=location,
        zenith_angle=zenith,
        azimuth_angle=azimuth,
        particle_altitude=palt,
        rigidity=rigidity_list[0],
        bfield_type=bfield_type,
        date=date_str,
    )

    max_step = int(np.ceil(max_time / dt))
    charge_si = traj.charge * ELEMENTARY_CHARGE
    mass_si = traj.mass * KG_PER_GEVC2

    global _worker_tracer
    if _worker_tracer is not None:
        # Process-level cache: update direction-specific start_altitude.
        tracer = _worker_tracer
        tracer.set_start_altitude(traj.start_alt)
    else:
        # Per-direction fallback: build one tracer for this direction
        # (loads IGRF once for the whole rigidity sweep).
        tracer = CppTrajectoryTracer(
            charge_si,
            mass_si,
            traj.start_alt,
            traj.esc_alt,
            dt,
            max_step,
            traj.bfield_type,
            traj.igrf_params,
            solver_char,
            atol,
            rtol,
        )

    # Precompute momentum direction unit vector (fixed across rigidities).
    ref_mom_si = traj.particle.momentum * KG_M_S_PER_GEVC
    pos = traj.particle_sixvector[:3]
    mom_unit = traj.particle_sixvector[3:] / ref_mom_si

    for rigidity in rigidity_list:
        traj.particle.set_from_rigidity(rigidity)
        mom_si = traj.particle.momentum * KG_M_S_PER_GEVC
        vec0 = list(pos) + list(mom_unit * mom_si)
        tracer.reset()
        tracer.evaluate(0.0, vec0)
        if tracer.particle_escaped:
            return (azimuth, zenith, rigidity)

    return (azimuth, zenith, 0.0)


def _evaluate_direction_cpp_only(args):
    """
    Lightweight thread worker for the tabulated IGRF case.
    All Python/numpy work (Trajectory, coordinate transforms) is done in the
    main thread; this function only builds a C++ tracer and calls
    find_cutoff_rigidity() which runs the entire rigidity loop in C++ with
    the GIL released.

    The tracer is cached per thread via threading.local() so that the
    expensive IGRF JSON parse (in the C++ constructor) happens only once
    per pool thread instead of once per direction.
    """
    (
        shared_table,
        table_params,
        igrf_params,
        charge_si,
        mass_si,
        start_alt,
        esc_alt,
        dt,
        max_step,
        solver_char,
        atol,
        rtol,
        pos,
        mom_unit,
        rigidity_list,
        mom_factor,
        azimuth,
        zenith,
    ) = args

    tracer = getattr(_thread_local, "tracer", None)
    if tracer is None:
        tracer = CppTrajectoryTracer(
            shared_table,
            table_params,
            charge_si,
            mass_si,
            start_alt,
            esc_alt,
            dt,
            max_step,
            igrf_params,
            solver_char,
            atol,
            rtol,
        )
        _thread_local.tracer = tracer
    else:
        tracer.set_start_altitude(start_alt)

    rcutoff = tracer.find_cutoff_rigidity(pos, mom_unit, rigidity_list, mom_factor)
    return (azimuth, zenith, rcutoff)


class GMRC:
    """
    Evaluates the geomagnetic cutoff rigidities associated to a specific location on the globe for each zenith and azimuthal angle (a zenith angle > 90 degrees are for upward-moving particles, that is, for cosmic rays coming from the other side of Earth).

    The cutoff rigidities are evaluated using a Monte-Carlo sampling scheme, combined with a 2-dimensional linear interpolation using `scipy.interpolate`.

    The resulting cutoffs can be plotted as 2-dimensional heatmap.

    Parameters
    -----------

    - location : str
        The location in which the geomagnetic cutoff rigidities are evaluated (default = "Kamioka"). The names must be one of the locations contained in `location_dict`, which is configured in `gtracr.utils`.
    - particle_altitude : float
        The altitude in which the cosmic ray interacts with the atmosphere in km (default = 100).
    - iter_num : int
        The number of iterations to perform for the Monte-Carlo sampling routine (default = 10000)
    - bfield_type : str
        The type of magnetic field model to use for the evaluation of the cutoff rigidities (default = "igrf"). Set to "dipole" to use the dipole approximation of the geomagnetic field instead.
    - particle_type : str
        The type of particle of the cosmic ray (default  ="p+").
    - date : str
        The specific date in which the geomagnetic rigidity cutoffs are evaluated. Defaults to the current date.
    - min_rigidity : float
        The minimum rigidity to which we evaluate the cutoff rigidities for (default = 5 GV).
    - max_rigidity : float
        The maximum rigidity to which we evaluate the cutoff rigidities for (default = 55 GV).
    - delta_rigidity : float
        The spacing between each rigidity (default = 5 GV). Sets the coarseness of the rigidity sample space.
    - n_workers : int, optional
        Number of parallel worker processes to use. Defaults to os.cpu_count()
        (all logical CPUs). Set to 1 to disable parallelism (useful for debugging).
    - solver : str, optional
        Integration method: "rk4" (frozen-field RK4, default), "boris" (Boris pusher),
        or "rk45" (adaptive Dormand-Prince).
    - atol, rtol : float, optional
        Absolute and relative tolerances for the RK45 adaptive solver (defaults 1e-3, 1e-6).

    Examples
    --------
    Python-orchestrated evaluation:

    >>> gmrc = GMRC(location="Kamioka", iter_num=1000, bfield_type="igrf")
    >>> gmrc.evaluate(dt=1e-5, max_time=1.)
    >>> az, zen, cutoffs = gmrc.interpolate_results()

    C++ batch mode (fastest):

    >>> gmrc = GMRC(location="Kamioka", iter_num=10000, bfield_type="table",
    ...             solver="rk45")
    >>> gmrc.evaluate_batch(dt=1e-5, max_time=1.)
    >>> az, zen, cutoffs = gmrc.bin_results()
    """

    def __init__(
        self,
        location="Kamioka",
        particle_altitude=100,
        iter_num=10000,
        bfield_type="igrf",
        particle_type="p+",
        date=str(date.today()),
        min_rigidity=5.0,
        max_rigidity=55.0,
        delta_rigidity=1.0,
        n_workers=None,
        solver="rk4",
        atol=1e-3,
        rtol=1e-6,
    ):
        # set class attributes
        self.location = location
        self.palt = particle_altitude
        self.iter_num = iter_num
        self.bfield_type = bfield_type
        self.plabel = particle_type
        self.date = date
        import os

        self.n_workers = n_workers if n_workers is not None else (os.cpu_count() or 1)
        self.solver_char = SOLVER_CHARS.get(solver.lower(), "r")
        self.atol = atol
        self.rtol = rtol
        # Rigidity configurations
        self.rmin = min_rigidity
        self.rmax = max_rigidity
        self.rdelta = delta_rigidity

        # generate list of rigidities
        self.rigidity_list = np.arange(self.rmin, self.rmax, self.rdelta)

        # initialize container for rigidity cutoffs
        self.data_dict = {
            "azimuth": np.zeros(self.iter_num),
            "zenith": np.zeros(self.iter_num),
            "rcutoff": np.zeros(self.iter_num),
        }

    def evaluate(self, dt=1e-5, max_time=1):
        """
        Evaluate the rigidity cutoff value at some provided location
        on Earth for a given cosmic ray particle.

        Uses parallel worker processes (one per CPU core by default) to
        evaluate independent Monte Carlo samples concurrently.

        Parameters
        ----------

        - dt : float
            The stepsize of each trajectory evaluation (default = 1e-5)
        - max_time : float
            The maximal time of each trajectory evaluation (default = 1.).

        """
        rigidity_list = list(self.rigidity_list)

        # Build argument list for each MC sample, using distinct random seeds
        # for reproducibility across different worker processes
        rng = np.random.default_rng()
        seeds = rng.integers(0, 2**31, size=self.iter_num)

        n_workers = self.n_workers
        use_parallel = n_workers > 1
        use_threads = use_parallel and self.bfield_type[0] == "t"

        if use_threads:
            # Tabulated IGRF: generate table once, share across threads.
            # All Python/numpy work (Trajectory construction, coordinate
            # transforms) is done here in the main thread so the thread
            # workers only execute GIL-released C++ code.
            from gtracr.utils import ymd_to_dec

            datapath = str(_DATA_DIR.resolve())
            dec_date = float(ymd_to_dec(self.date))
            igrf_params = (datapath, dec_date)
            shared_table, table_params = _get_or_generate_igrf_table(datapath, dec_date)

            max_step = int(np.ceil(max_time / dt))

            # Pre-compute initial conditions for every MC direction (main thread).
            args_list = []
            for i in range(self.iter_num):
                rng_i = np.random.default_rng(int(seeds[i]))
                azimuth, zenith = rng_i.random(2) * np.array([360.0, 180.0])

                traj = Trajectory(
                    plabel=self.plabel,
                    location_name=self.location,
                    zenith_angle=zenith,
                    azimuth_angle=azimuth,
                    particle_altitude=self.palt,
                    rigidity=rigidity_list[0],
                    bfield_type="igrf",  # only for coordinate transform
                    date=self.date,
                )

                charge_si = traj.charge * ELEMENTARY_CHARGE
                mass_si = traj.mass * KG_PER_GEVC2
                mom_factor = float(np.abs(traj.charge)) * KG_M_S_PER_GEVC
                ref_mom_si = traj.particle.momentum * KG_M_S_PER_GEVC
                pos = tuple(float(x) for x in traj.particle_sixvector[:3])
                mom_unit = tuple(
                    float(x) for x in traj.particle_sixvector[3:] / ref_mom_si
                )

                args_list.append(
                    (
                        shared_table,
                        table_params,
                        igrf_params,
                        charge_si,
                        mass_si,
                        traj.start_alt,
                        traj.esc_alt,
                        dt,
                        max_step,
                        self.solver_char,
                        self.atol,
                        self.rtol,
                        pos,
                        mom_unit,
                        rigidity_list,
                        mom_factor,
                        azimuth,
                        zenith,
                    )
                )

            results = []
            with ThreadPoolExecutor(max_workers=n_workers) as executor:
                future_to_idx = {
                    executor.submit(_evaluate_direction_cpp_only, args): i
                    for i, args in enumerate(args_list)
                }
                for future in tqdm(
                    as_completed(future_to_idx),
                    total=self.iter_num,
                    desc="GMRC evaluation (threaded)",
                ):
                    i = future_to_idx[future]
                    results.append((i, future.result()))

            for i, (az, zen, rc) in results:
                self.data_dict["azimuth"][i] = az
                self.data_dict["zenith"][i] = zen
                self.data_dict["rcutoff"][i] = rc
        else:
            args_list = [
                (
                    self.location,
                    self.plabel,
                    self.bfield_type,
                    self.date,
                    self.palt,
                    rigidity_list,
                    dt,
                    max_time,
                    int(seeds[i]),
                    self.solver_char,
                    self.atol,
                    self.rtol,
                )
                for i in range(self.iter_num)
            ]

            if use_parallel:
                # Parallel evaluation: each worker process evaluates one MC sample.
                # The initializer creates one cached CppTrajectoryTracer per worker,
                # eliminating repeated IGRF construction inside each task.
                common_params = (
                    self.plabel,
                    self.bfield_type[0],
                    self.date,
                    self.palt,
                    dt,
                    max_time,
                    self.solver_char,
                    self.atol,
                    self.rtol,
                )
                results = []
                with ProcessPoolExecutor(
                    max_workers=n_workers,
                    initializer=_init_worker,
                    initargs=(common_params,),
                ) as executor:
                    future_to_idx = {
                        executor.submit(_evaluate_single_direction, args): i
                        for i, args in enumerate(args_list)
                    }
                    for future in tqdm(
                        as_completed(future_to_idx),
                        total=self.iter_num,
                        desc="GMRC evaluation",
                    ):
                        i = future_to_idx[future]
                        results.append((i, future.result()))

                for i, (az, zen, rc) in results:
                    self.data_dict["azimuth"][i] = az
                    self.data_dict["zenith"][i] = zen
                    self.data_dict["rcutoff"][i] = rc
            else:
                # Sequential fallback (n_workers=1), useful for debugging
                for i in tqdm(range(self.iter_num)):
                    az, zen, rc = _evaluate_single_direction(args_list[i])
                    self.data_dict["azimuth"][i] = az
                    self.data_dict["zenith"][i] = zen
                    self.data_dict["rcutoff"][i] = rc

    def evaluate_batch(self, dt=1e-5, max_time=1.0, base_seed=None):
        """
        Evaluate rigidity cutoffs entirely in C++ (batch mode).

        The entire MC loop — RNG, coordinate transforms, rigidity scanning,
        and threading — runs in a single C++ call, eliminating all Python
        overhead per ray.

        Parameters
        ----------
        dt : float
            Step size for trajectory integration (default 1e-5 s).
        max_time : float
            Maximum integration time per trajectory (default 1 s).
        base_seed : int, optional
            RNG seed for reproducibility. If None, a random seed is used.
        """
        import time as _time

        from gtracr._libgtracr import BatchGMRCParams, TableParams
        from gtracr._libgtracr import batch_gmrc_evaluate as _batch_eval
        from gtracr.utils import location_dict, particle_dict, ymd_to_dec

        t_wall_start = _time.monotonic()

        use_table = self.bfield_type[0] == "t"
        field_label = "table" if use_table else "direct IGRF"
        print(
            f"Initializing batch GMRC for {self.location} "
            f"({self.iter_num} samples, {self.n_workers} threads, "
            f"{field_label})..."
        )

        datapath = str(_DATA_DIR.resolve())
        dec_date = float(ymd_to_dec(self.date))
        igrf_params = (datapath, dec_date)

        if use_table:
            shared_table, table_params = _get_or_generate_igrf_table(datapath, dec_date)
        else:
            shared_table = None
            table_params = TableParams()

        loc = location_dict[self.location]
        particle = particle_dict[self.plabel]

        if base_seed is None:
            base_seed = int(np.random.default_rng().integers(0, 2**63))

        p = BatchGMRCParams()
        p.latitude = loc.latitude
        p.longitude = loc.longitude
        p.detector_alt = loc.altitude  # km
        p.particle_alt = self.palt  # km (GMRC stores km)
        p.escape_radius = 10.0 * EARTH_RADIUS
        p.charge = float(particle.charge)  # units of e
        p.mass = particle.mass  # GeV/c^2
        p.min_rigidity = self.rmin
        p.max_rigidity = self.rmax
        p.delta_rigidity = self.rdelta
        p.dt = dt
        p.max_time = max_time
        p.solver_type = self.solver_char
        p.bfield_type = self.bfield_type[0]
        p.atol = self.atol
        p.rtol = self.rtol
        p.n_samples = self.iter_num
        p.n_threads = self.n_workers
        p.max_attempts_factor = 30
        p.base_seed = base_seed

        t_init_done = _time.monotonic()
        print(
            f"Initialized in {t_init_done - t_wall_start:.2f}s. "
            f"Running cutoff calculations..."
        )

        zenith, azimuth, rcutoff, total_traj = _batch_eval(
            shared_table, table_params, igrf_params, p
        )

        t_calc_done = _time.monotonic()
        n = len(zenith)
        calc_elapsed = t_calc_done - t_init_done
        total_elapsed = t_calc_done - t_wall_start
        ktraj_per_s = (
            (total_traj / 1000.0) / calc_elapsed if calc_elapsed > 0 else float("inf")
        )

        print(
            f"Completed {n} cutoffs in {calc_elapsed:.2f}s "
            f"({total_traj} trajectories, {ktraj_per_s:.1f}k traj/s "
            f"across {self.n_workers} threads, {total_elapsed:.2f}s total)"
        )

        # Store results (may be shorter than iter_num if safety limit hit)
        self.data_dict = {
            "azimuth": azimuth,
            "zenith": zenith,
            "rcutoff": rcutoff,
        }
        if n < self.iter_num:
            import warnings

            warnings.warn(
                f"Batch GMRC: only {n}/{self.iter_num} successful samples "
                f"(safety limit reached). Results may be sparse."
            )

    def interpolate_results(self, method="linear", ngrid_azimuth=70, ngrid_zenith=70):
        """
        Interpolate the rigidity cutoffs using `scipy.interpolate.griddata`

        Legacy method kept for backward compatibility; prefer
        ``bin_results`` for large sample counts.

        Parameters
        ----------
        - method : str
            The type of linear interpolation used for `griddata` (default = "linear"). Choices are between "nearest", "linear", and "cubic".
        - ngrid_azimuth, ngrid_zenith : int
            The number of grids for the azimuth and zenith angles used for the interpolation (default = 70).

        Returns
        --------

        Returns a tuple of the following objects:

        - azimuth_grid : np.array(float), size ngrid_azimuth
            The linearly spaced values of the azimuthal angle
        - zenith_grid : np.array(float), size ngrid_zenith
            The linearly spaced values of the zenith angle
        - rcutoff_grid : np.array(float), size ngrid_azimuth x ngrid_zenith
            The interpolated geomagnetic cutoff rigidities.
        """

        azimuth_grid = np.linspace(
            np.min(self.data_dict["azimuth"]),
            np.max(self.data_dict["azimuth"]),
            ngrid_azimuth,
        )
        zenith_grid = np.linspace(
            np.max(self.data_dict["zenith"]),
            np.min(self.data_dict["zenith"]),
            ngrid_zenith,
        )

        rcutoff_grid = griddata(
            points=(self.data_dict["azimuth"], self.data_dict["zenith"]),
            values=self.data_dict["rcutoff"],
            xi=(azimuth_grid[None, :], zenith_grid[:, None]),
            method=method,
        )

        return (azimuth_grid, zenith_grid, rcutoff_grid)

    def bin_results(self, nbins_azimuth=72, nbins_zenith=36):
        """
        Bin the MC samples into a regular (azimuth, zenith) grid and compute
        the mean cutoff rigidity per bin.  Much faster than scattered
        interpolation and more appropriate when the sample count is large.

        Parameters
        ----------
        nbins_azimuth : int
            Number of azimuth bins (default 72, i.e. 5-degree bins).
        nbins_zenith : int
            Number of zenith bins (default 36, i.e. 5-degree bins).

        Returns
        -------
        (azimuth_centres, zenith_centres, rcutoff_grid) where
        azimuth_centres and zenith_centres are 1-D bin-centre arrays and
        rcutoff_grid is shape (nbins_zenith, nbins_azimuth).  Bins with
        no samples are filled with NaN.
        """
        az = self.data_dict["azimuth"]
        zen = self.data_dict["zenith"]
        rc = self.data_dict["rcutoff"]

        az_edges = np.linspace(0.0, 360.0, nbins_azimuth + 1)
        zen_edges = np.linspace(0.0, 180.0, nbins_zenith + 1)

        # Digitize: bin index 1..N for in-range, 0 or N+1 for out-of-range
        az_idx = np.digitize(az, az_edges) - 1  # 0-based
        zen_idx = np.digitize(zen, zen_edges) - 1

        # Clamp to valid range
        az_idx = np.clip(az_idx, 0, nbins_azimuth - 1)
        zen_idx = np.clip(zen_idx, 0, nbins_zenith - 1)

        # Accumulate sum and count per bin
        sum_grid = np.zeros((nbins_zenith, nbins_azimuth))
        count_grid = np.zeros((nbins_zenith, nbins_azimuth))
        np.add.at(sum_grid, (zen_idx, az_idx), rc)
        np.add.at(count_grid, (zen_idx, az_idx), 1)

        with np.errstate(invalid="ignore"):
            rcutoff_grid = sum_grid / count_grid  # NaN where count == 0

        az_centres = 0.5 * (az_edges[:-1] + az_edges[1:])
        zen_centres = 0.5 * (zen_edges[:-1] + zen_edges[1:])

        return (az_centres, zen_centres, rcutoff_grid)



"""
Evaluate and plot a single cosmic ray trajectory.

Supports interactive display or file output.
"""

import argparse
from pathlib import Path

import numpy as np

from gtracr.constants import EARTH_RADIUS
from gtracr.plotting import plot_2dtraj, plot_3dtraj, plot_traj_momentum
from gtracr.trajectory import Trajectory
from gtracr.utils import dec_to_dms

PLOT_DIR = Path(__file__).parent.parent.parent / "gtracr_plots"


def convert_to_cartesian(trajectory_data):
    r_arr = trajectory_data["r"] / EARTH_RADIUS
    theta_arr = trajectory_data["theta"]
    phi_arr = trajectory_data["phi"]

    trajectory_data["x"] = r_arr * np.sin(theta_arr) * np.cos(phi_arr)
    trajectory_data["y"] = r_arr * np.sin(theta_arr) * np.sin(phi_arr)
    trajectory_data["z"] = r_arr * np.cos(theta_arr)


def plot_trajectory(traj_datadict, title, check_3dtraj=False, show_plot=False):
    convert_to_cartesian(traj_datadict)
    plot_2dtraj([traj_datadict], plotdir_path=str(PLOT_DIR), show_plot=show_plot)
    if check_3dtraj:
        plot_3dtraj(
            [traj_datadict],
            title_name=title,
            plotdir_path=str(PLOT_DIR),
            show_plot=show_plot,
        )


def run(args):
    PLOT_DIR.mkdir(parents=True, exist_ok=True)

    traj = Trajectory(
        plabel=args.particle,
        location_name=args.location,
        zenith_angle=args.zenith,
        azimuth_angle=args.azimuth,
        rigidity=args.rigidity,
        bfield_type=args.bfield_type,
        solver=args.solver,
    )

    if args.bfield_type == "table":
        print("Building IGRF lookup table (64×128×256 grid)…", flush=True)

    traj_datadict = traj.get_trajectory(
        dt=args.dt, max_time=args.max_time, get_data=True, use_python=False
    )

    if args.bfield_type == "table":
        print("Done.", flush=True)

    steps = len(traj_datadict["t"])
    print(f"particle_escaped = {traj.particle_escaped}  steps = {steps}")

    lat = traj.latitude
    lng = traj.longitude
    lat_dms, lng_dms = dec_to_dms(lat, lng)

    title = (
        f"Particle Trajectory at {lat_dms:s}, {lng_dms:s} "
        f"with Zenith {args.zenith:.1f}°, Azimuth {args.azimuth:.1f}°, "
        f"R = {args.rigidity:.1f} GV [{args.bfield_type}]"
    )

    plot_traj_momentum(traj_datadict, args.rigidity, args.show_plot)
    plot_trajectory(traj_datadict, title, check_3dtraj=True, show_plot=args.show_plot)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Evaluate and plot a single cosmic ray trajectory."
    )
    parser.add_argument(
        "--location",
        default="Kamioka",
        help="Detector location name (default: Kamioka).",
    )
    parser.add_argument(
        "--particle",
        default="p+",
        choices=["p+", "p-", "e+", "e-"],
        help="Particle label (default: p+).",
    )
    parser.add_argument(
        "--rigidity",
        type=float,
        default=10.0,
        help="Particle rigidity in GV (default: 10.0).",
    )
    parser.add_argument(
        "--zenith",
        type=float,
        default=0.0,
        help="Zenith angle in degrees (default: 0.0).",
    )
    parser.add_argument(
        "--azimuth",
        type=float,
        default=0.0,
        help="Azimuth angle in degrees (default: 0.0).",
    )
    parser.add_argument(
        "--solver",
        default="rk4",
        choices=["rk4", "boris", "rk45"],
        help="Integration method: rk4 (default), boris, or rk45 (adaptive).",
    )
    parser.add_argument(
        "--bfield-type",
        dest="bfield_type",
        default="igrf",
        choices=["igrf", "dipole", "table"],
        help="Magnetic field model (default: igrf).",
    )
    parser.add_argument(
        "--dt",
        type=float,
        default=1e-5,
        help="Integration step size in seconds (default: 1e-5).",
    )
    parser.add_argument(
        "--max-time",
        dest="max_time",
        type=float,
        default=1.0,
        help="Maximum integration time in seconds (default: 1.0).",
    )
    parser.add_argument(
        "--show-plot",
        dest="show_plot",
        action="store_true",
        help="Show plots in an interactive window.",
    )
    args = parser.parse_args()
    run(args)



"""Geographic detector location data class."""


class Location:
    """
    Geographic detector location.

    Parameters
    ----------
    name : str
        Human-readable location name (e.g. ``"Kamioka"``).
    latitude : float
        Geographic latitude in decimal degrees (positive = north).
    longitude : float
        Geographic longitude in decimal degrees (positive = east).
    altitude : float, optional
        Altitude above sea level in km (default 0).
    """

    def __init__(self, name, latitude, longitude, altitude=0.0):
        self.name = name
        self.latitude = latitude
        self.longitude = longitude
        self.altitude = altitude

    def __str__(self):
        return (
            f"{self.name}: latitude={self.latitude}, "
            f"longitude={self.longitude}, altitude={self.altitude} km"
        )



"""
Utility functions for gtracr.

Provides coordinate conversions (decimal ↔ DMS, decimal year), a pickled
dictionary loader, and factory functions for the predefined location and
particle dictionaries.
"""

import pickle
from pathlib import Path

import numpy as np

from gtracr.location import Location
from gtracr.particle import Particle

_DATA_DIR = Path(__file__).parent / "data"


def dec_to_dms(lat_dec, lng_dec):
    """
    Converts geodesic coordinates expressed in decimal notation
    to DMS notation (degree, minutes, seconds).

    Parameters
    ----------
    - lat_dec : float
        the latitude in decimal notation ([-90, 90])
    - lng_dec : float
        the longitude in decimal notation ([-180,180])

    Returns
    --------
    - lat_dms : str
        latitude in DMS notation
    - lng_dms : str
        longitude in DMS notation

    The evaluation is performed with reference to the DMS to decimal calculator:
    https://www.rapidtables.com/convert/number/degrees-to-degrees-minutes-seconds.html

    The convention follows the WGS 84 convention as posed here
    : https://gisgeography.com/latitude-longitude-coordinates/

    The convention shows that:
    - A positive latitude (latitude > 0) is towards the North Pole
    - A negative latitude (latitude < 0) is towards the South Pole
    - A positive longitude (longitude > 0) is towards the East relative to one's location
    - A negative longitude (longitude < 0) is towards the West relative to one's location

    So with this, we see that:
    - latitude goes counter-clockwise from the equator
    - longitude goes counter-clockwise from the Prime Meridian
    """
    # latitude in dms
    lat_deg = int(np.floor(lat_dec))
    lat_min = int(np.floor((lat_dec - lat_deg) * 60.0))
    lat_sec = int(np.floor((lat_dec - lat_deg - (lat_min / 60.0)) * 60.0))

    # add north or south depending on sign of lat_dec
    lat_symb = "N" if lat_dec >= 0 else "S"

    lat_dms = f"{lat_deg:d}°{lat_min:d}'{lat_sec:d}\"{lat_symb:s}"

    # longitude in dms
    lng_deg = int(np.floor(lng_dec))
    lng_min = int(np.floor((lng_dec - lng_deg) * 60.0))
    lng_sec = int(np.floor((lng_dec - lng_deg - (lng_min / 60.0)) * 60.0))

    # add east or west depending on sign of lng_dec
    lng_symb = "E" if lng_dec >= 0 else "W"

    lng_dms = f"{lng_deg:d}°{lng_min:d}'{lng_sec:d}\"{lng_symb:s}"

    return lat_dms, lng_dms


def ymd_to_dec(ymd_date):
    """
    Converts a date in yyyy-mm-dd format into decimal format in
    units of years.

    Parameters
    ----------

    - ymd : str
        the year, month, and date of the specified date in yyyy-mm-dd
        format.

    Returns
    ---------

    - dec_date : float
        the date in decimal format, in units of years.

    """
    # break down the str to get year, month, date separately
    year, month, day = [float(val) for val in ymd_date.split("-")]

    # the number of days in each month
    # not considering leap years right now
    days_per_mth = np.array(
        [31.0, 28.0, 31.0, 30.0, 31.0, 30.0, 31.0, 31.0, 30.0, 31.0, 30.0, 31.0]
    )

    # get the total number of days based on the month + day
    ndays = float(days_per_mth[int(month) - 1]) + day

    # check if year is a leap year or not
    # this will change the maximum # days in a year
    # also will change the number of days to the given date
    max_ndays = 365
    if year % 4 == 0:
        max_ndays += 1
        ndays += 1

    # convert month into decimal years
    dec_mth = month / 12.0

    # convert days into decimal years
    dec_days = ndays / max_ndays

    # return the sum of year, month, day
    return year + dec_mth + dec_days


def import_dict(fname):
    """
    Load a pickled dictionary from disk.

    Parameters
    ----------
    fname : str
        Absolute or relative path to the ``.pkl`` file.

    Returns
    -------
    dict
        The deserialized dictionary.
    """
    with open(fname, "rb") as f:
        the_dict = pickle.load(f)
    return the_dict


def set_locationdict():
    """
    Build the dictionary of predefined detector locations.

    Returns
    -------
    dict of str -> Location
        Mapping from location name to ``Location`` instance.
        Predefined sites: Kamioka, IceCube, SNOLAB, UofA,
        CTA-North, CTA-South, ORCA, ANTARES, Baikal-GVD, TA.
    """
    location_dict = {}

    locations = [
        Location("Kamioka", 36.434800, 137.276599, 0.0),
        Location("IceCube", -89.99, 0.0, 0.0),
        Location("SNOLAB", 46.471983, -81.186701, 0.0),
        Location("UofA", 53.523230, -113.526319, 0.0),
        Location("CTA-North", 28.76216, -17.89201, 0.0),
        Location("CTA-South", -24.68342778, -24.68342778, 0.0),
        Location("ORCA", 42.80000000, 6.0333333, 0.0),
        Location("ANTARES", 42.8, 6.1666666, 0.0),
        Location("Baikal-GVD", 51.77139, 104.3978, 0.0),
        Location("TA", 39.208099, -113.121285, 0.0),
    ]

    # add location to location_dict if it does not exist
    for loc in locations:
        if loc.name not in list(location_dict.keys()):
            location_dict[loc.name] = loc
        else:
            continue

    return location_dict


def set_particledict():
    """
    Build the dictionary of supported cosmic ray particles.

    Returns
    -------
    dict of str -> Particle
        Mapping from particle label to ``Particle`` instance.
        Supported labels: ``"p+"``, ``"p-"``, ``"e+"``, ``"e-"``.
    """
    particle_dict = {}

    # different cosmic ray particles
    particles = [
        Particle("positron", -11, 0.5109 * (1e-3), 1, "e+"),
        Particle("electron", 11, 0.5109 * (1e-3), -1, "e-"),
        Particle("proton", 2212, 0.937272, 1, "p+"),
        Particle("anti-proton", -2212, 0.937272, -1, "p-"),
    ]

    # add particle to particle_dict if it does not exist
    for particle in particles:
        if particle.name not in list(particle_dict.keys()):
            particle_dict[particle.label] = particle
        else:
            continue

    return particle_dict


location_dict = set_locationdict()
particle_dict = set_particledict()



"""Cosmic ray particle data class."""

import numpy as np


class Particle:
    """
    Cosmic ray particle with kinematic properties.

    Parameters
    ----------
    name : str
        Human-readable particle name (e.g. ``"proton"``).
    pid : int
        PDG Monte Carlo particle ID.
    mass : float
        Rest mass in GeV/c².
    charge : int
        Electric charge in units of elementary charge *e*.
    label : str
        Short label used as dictionary key (e.g. ``"p+"``).

    Notes
    -----
    PDG IDs: http://pdg.lbl.gov/2007/reviews/montecarlorpp.pdf

    Examples
    --------
    >>> proton = Particle("proton", 2212, 0.938272, 1, "p+")
    """

    def __init__(self, name, pid, mass, charge, label):
        self.name = name
        self.pid = pid
        self.mass = mass
        self.charge = charge
        self.label = label
        self.momentum = 0.0
        self.rigidity = 0.0

    def set_from_energy(self, energy):
        """Set momentum and rigidity from total relativistic energy.

        Parameters
        ----------
        energy : float
            Total relativistic energy in GeV.
        """
        self.momentum = np.sqrt(energy**2.0 - self.mass**2.0)
        self.rigidity = self.momentum / np.abs(self.charge)

    def set_from_rigidity(self, rigidity):
        """Set momentum and rigidity from magnetic rigidity.

        Parameters
        ----------
        rigidity : float
            Magnetic rigidity in GV (momentum / charge).
        """
        self.momentum = rigidity * np.abs(self.charge)
        self.rigidity = rigidity

    def set_from_momentum(self, momentum):
        """Set momentum and compute rigidity from relativistic momentum.

        Parameters
        ----------
        momentum : float
            Relativistic momentum in GeV/c.
        """
        self.momentum = momentum
        self.rigidity = self.momentum / np.abs(self.charge)

    def get_energy_rigidity(self):
        """Compute total energy from the current rigidity.

        Returns
        -------
        float
            Total relativistic energy in GeV.
        """
        return (
            np.sqrt((self.rigidity * np.abs(self.charge)) ** 2.0 + self.mass**2.0)
            + self.mass
        )

    def __str__(self):
        return (
            f"{self.name}: PID = {self.pid}, m = {self.mass:.6f} GeV, "
            f"Z = {self.charge}e\n"
            f"  Momentum = {self.momentum:.6e}, Rigidity = {self.rigidity:.6e}"
        )



"""
Magnetic field models for Earth's geomagnetic field.

Classes
-------
MagneticField : ideal dipole approximation.
IGRF13 : IGRF-13 spherical harmonic model (degree 13, 1900–2025).
IGRFTable : pre-computed 3-D lookup table with trilinear interpolation.
"""

from gtracr.bfield.dipole import MagneticField
from gtracr.bfield.igrf import IGRF13
from gtracr.bfield.table import IGRFTable

__all__ = ["MagneticField", "IGRF13", "IGRFTable"]



"""Physical constants and unit conversions for gtracr."""

import numpy as np
import scipy.constants as sc

SPEED_OF_LIGHT = sc.c  # speed of light in vacuum (m/s)

ELEMENTARY_CHARGE = 1.602e-19  # elementary charge (coulombs)

EARTH_RADIUS = 6371.2 * (1e3)  # Earth radius (meters)
G10 = 29404.8 * (1e-9)  # IGRF-2020 first dipole coefficient (Tesla)

RAD_PER_DEG = np.pi / 180.0
DEG_PER_RAD = 180.0 / np.pi

KG_PER_GEVC2 = 1.78e-27  # kg per GeV/c²
KG_M_S_PER_GEVC = 5.36e-19  # kg·m/s per GeV/c

# Solver name → C++ single-character mapping (shared by Trajectory and GMRC)
SOLVER_CHARS: dict[str, str] = {"rk4": "r", "boris": "b", "rk45": "a"}
